require 'bdd'

class VariableSet
  attr_reader :from, :to, :bdd_from, :bdd_to, :n

  def initialize(n)
    BDD.set_var_num(2*n)
    @n = n
    @from = (0..n-1).collect { |x| BDD.get_var(2*x) }
    @to = (0..n-1).collect { |x| BDD.get_var(2*x+1) }
    @bdd_from = BDD.make_set(@from)
    @bdd_to = BDD.make_set(@to)
    @succ = BDD::Pair.new
    @pred = BDD::Pair.new
    0.upto(n-1) { |i|
      @succ << [ @to[i], @from[i] ]
      @pred << [ @from[i], @to[i] ]
    }
  end

  def current2next
    @succ
  end

  def next2current
    @pred
  end

  def assigned(vars)
    res = BDD.true
    (@n-1).downto(0) { |i|
      res &= (@from[i] <=> @to[i]) if !vars.include?(i)
    }
    res
  end
end

def reachable_states(init,trans,var)
  r1 = init
  r2 = BDD.false
  until r1 == r2
    r2 = r1
#   r1 |= (r1 & trans).exist(var.bdd_from).replace(var.current2next)
    r1 |= r1.apply_exist(trans, BDD::AND, var.bdd_from).replace(var.current2next)
  end
  r1
end

#############

class Milner
  attr_reader :var, :trans, :init

  def initialize(n)
    @var = VariableSet.new(3*n) # ci, ti, hi
    @trans = BDD.false
    @init = BDD.true
    @n = n
    0.upto(n-1) { |i|
      pi = c1(i) & t1(i).not & t2(i) & c2(i).not & h2(i) & assigned(c(i),t(i),h(i))
      pi |= h1(i) & c2((i+1)%n) & h2(i).not & assigned(c((i+1)%n),h(i))
      ei = t1(i) & t2(i).not & assigned(t(i))
      @trans |= pi | ei
      @init &= t1(i).not & h1(i).not & (i == 0 ? c1(i) : c1(i).not)
    }

    # Name variables - only for printing
    0.upto(n-1) { |i|
      eval("BDD.var('c_#{i}')")
      eval("BDD.var('c\\'_#{i}')")
    }
    0.upto(n-1) { |i|
      eval("BDD.var('t_#{i}')")
      eval("BDD.var('t\\'_#{i}')")
    }
    0.upto(n-1) { |i|
      eval("BDD.var('h_#{i}')")
      eval("BDD.var('h\\'_#{i}')")
    }
  end

  def one_token
    BDD.nxor((0..@n-1).collect{|i| c1(i)} + (0..@n-1).collect{|i| h1(i)})
  end

  def c(i)
    i
  end
  def t(i)
    i+@n
  end
  def h(i)
    i+2*@n
  end

  def c1(i)
    @var.from[c(i)]
  end
  def t1(i)
    @var.from[t(i)]
  end
  def h1(i)
    @var.from[h(i)]
  end

  def c2(i)
    @var.to[c(i)]
  end
  def t2(i)
    @var.to[t(i)]
  end
  def h2(i)
    @var.to[h(i)]
  end

  def assigned(*v)
    @var.assigned(v)
  end

  def states
    return @reachable if @reachable
    @reachable = reachable_states(@init, @trans, @var)
  end

  def deadlocks
    return @deadlocks if @deadlocks
    @deadlocks = states & @trans.exist(@var.bdd_to).not
  end
end

m = Milner.new(4)
#m.init.show
#m.states.show
check = m.states.imply(m.one_token)
#check.show
puts "Only one token: " + check.true?.to_s
#m.deadlocks.show
puts "Deadlocks: " + (m.deadlocks.false? ? "no" : "yes")
m.deadlocks.each{ |x| puts x.to_keys.inspect }
